(* uses src/missing,
        src/parser,
        src/psi-parsing
 *)


signature AST =
sig
  datatype ast = Node of string * (ast list)

  structure Parser : PARSER
  val parser : string -> ast Parser.parser
end;


functor Ast(S: PARSER_STREAM) (* : AST *) =
struct
  structure Parser = Parser(S)

  open Missing

  datatype ast = Node of string * (ast list)

  fun node typ nodes = Node (typ,nodes)

  (*val x = Node ("if", [Node ("ifExp", []), Node ("thenExp", [])])*)


  structure P = Parser(StringStream)
  structure L = PsiLexerParserComb(P)

  datatype p_seq_i = PStr of string | PCl of string
  type p_seq = p_seq_i list
  datatype p_clause = PClause of string * (p_seq list)


  local open P open L in


  val sp = sat Char.isSpace
  val sps = many1 sp

  fun clauseName () =
                        stok "<" >>
                        identifier >>=
    (fn clName =>       stok ">" >>
                        return clName)


  fun notIt () = 
    ((stok "|||" >> return []) 
              </choice/> (eof [])
              </choice/> (sp >> return []) )

  fun parseIt () =
    (sps >> eof () >> error "End of input unexpected") </choice/>
    (clauseName () >>= return o PCl)
    </choice/>
    (ws >>
    ((until  )
        >>= return o PStr o String.implode o reverse))


  fun parseSeq () =
    many1 $ parseIt ()

  fun parseClause () =
                        clauseName () >>=
    (fn clName =>       stok "::=" >>
                        (parseSeq () </sepby1/> stok "|||") >>=
    (fn seq =>          return $ PClause (clName, seq) ))


  fun test () =
  let
    val s = "<IfExp> ::= if ( <exp> ) { <exp> } else { <exp> } ||| if ( <exp> ) <exp> "
    (*val s = "if  |||"*)
  in
    PolyML.print (parse (parseClause ()) (StringStream.make s))
  end
  end

  fun parser s = Parser.return (Node ("", []))

end;



structure AstTest = Ast(StringStream);
PolyML.print_depth 100;
AstTest.test ();

(*


if ( <exp> )

<IfExp> ::= if ( exp ) { exp } else { exp }
        ||| if (exp) exp

<exp> ::= <integer>, <name>, 

<IfElseExp> ::= if ( <exp> ) { <exp> } else { <exp> }
             |  if ( <exp> ) <exp> else <exp>

 *)
