(*
    uses src/missing,
         src/parser
*)

structure Uses :
sig
  exception UsesError of string

  val uses : string -> unit
  val usesFile : string -> unit
  val collectFileListFromModStrFiles : string -> string list
end =
struct
  structure Parser = Parser(ParserStreamIO)
  open Parser
  open Missing

  exception UsesError of string

  val home = PsiWorkBenchHome.home

  fun modStrToFileName modStr =
    if String.isPrefix "@" modStr then
      String.extract (modStr,1,NONE)
    else
      home ^ modStr ^ ".ML"


  fun isWord chr  = Char.isGraph chr andalso (chr <> #",")
  fun isSpace chr = Char.isSpace chr
  fun toStr l     = String.concat (map Char.toString l)
  val ws          = sat isSpace
  fun word ()     = many1 (sat isWord) >>= (fn l => return (toStr l))


  (*
    <uses-clause> ::= `uses' <word> (`,' <word>)*
  *)
  fun useParser () =
                    str "uses" >> nofail (many1 ws >>
                    ((many ws >> word ()) </sepby1/> (many ws >> chr #",")))

  (*
    <uses-pragma> ::= `(*' <uses-clause> `*)'
  *)
  fun usesParser () =
                    many ws >> str "(*" >> many ws >>
                    useParser () >>=
    (fn uses =>     nofail (many ws >> str "*)" >>
                    return uses))


  fun readUsesFromFile filename =
  let
    val fh    = TextIO.openIn filename
    val fs    = ParserStreamIO.make fh
    val files = parse (usesParser ()) fs
    val ()    = TextIO.closeIn fh
  in
    case files of
         LEFT ((Parser.Error msg,_)::_) => raise UsesError msg
       | RIGHT ((fs,_)::_)              => fs
       | _                              => []
  end


  fun collectFileListFromModStrFiles modStr =
  let
    val loaded = ref []

    fun loadFile modStr visited =
      if Lst.member modStr visited then raise UsesError "circular" else
      if Lst.member modStr (!loaded) then () else
        let
          val fileName = modStrToFileName modStr
          val modStrs = readUsesFromFile fileName
          val visited = modStr :: visited
          val () = forall (fn ms => loadFile ms visited) modStrs
          val () = loaded := (!loaded) @ [modStr]
        in
          ()
        end

    val () = loadFile modStr []
  in
    !loaded
  end


  fun printErr str = TextIO.output (TextIO.stdErr, str)

  fun uses' modStr =
  let
    val files = collectFileListFromModStrFiles modStr
  in
    forall use (map (fn f => modStrToFileName f) files)
  end

  fun uses modStr =
    uses' modStr
        handle (IO.Io {name, ...}) =>
            printErr ("\nError: could not read file \"" ^ name ^ "\"\n")

  fun usesFile filename = uses ("@" ^ filename)

end;

